module Parsing where

import Char

-------------------------------------------------------------------------------
-- parser type
-------------------------------------------------------------------------------

type Parser a = String -> [(a, String)]

-------------------------------------------------------------------------------
-- elementaire parsers
-------------------------------------------------------------------------------

succeed      :: a -> Parser a
succeed x xs =  [(x, xs)] 

satisfy                        :: (Char -> Bool) -> Parser Char
satisfy _ []                   =  []
satisfy p (x : xs) | p x       =  [(x, xs)]
                   | otherwise =  []

-------------------------------------------------------------------------------
-- combinators
-------------------------------------------------------------------------------

infixl 5 <*>
infixr 3 <|>

(<*>)        :: Parser (a -> b) -> Parser a -> Parser b
(p <*> q) xs =  [(f x, xs'') | (f, xs') <- p xs, (x, xs'') <- q xs']

(<|>)        :: Parser a -> Parser a -> Parser a
(p <|> q) xs =  p xs ++ q xs

-------------------------------------------------------------------------------
-- afgeleide parsers
-------------------------------------------------------------------------------

letter        :: Char -> Parser Char
letter      c =  satisfy (== c)

anythingBut    :: [Char] -> Parser Char
anythingBut cs =  satisfy (`notElem` cs)

tok :: String -> Parser String
tok =  foldr op (succeed [])
  where
    op c p = (:) <$> letter c <*> p

-------------------------------------------------------------------------------
-- afgeleide combinators
-------------------------------------------------------------------------------

infixl 5 <* ,  *>
infix  7 <$>, <$

(<$>)   :: (a -> b) -> Parser a -> Parser b
f <$> p =  succeed f <*> p

(<$)   :: b -> Parser a -> Parser b
y <$ p =  const y <$> p

(<*)   :: Parser b -> Parser a -> Parser b
p <* q =  const <$> p <*> q

(*>)   :: Parser b -> Parser a -> Parser a
p *> q =  id <$ p <*> q

-------------------------------------------------------------------------------
-- parsers voor lijsten
-------------------------------------------------------------------------------

many   :: Parser a -> Parser [a]
many p =  (:) <$> p <*> many p <|> succeed []

many1   :: Parser a -> Parser [a]
many1 p =  (:) <$> p <*> many p

exactly                 :: Int -> Parser a -> Parser [a]
exactly n p | n <= 0    =  succeed []
            | otherwise =  (:) <$> p <*> exactly (n - 1) p

-------------------------------------------------------------------------------
-- cijfers
-------------------------------------------------------------------------------

digit :: Parser Char
digit =  satisfy isDigit

natural :: Parser Int
natural =  int <$> many1 digit

natural_   :: Int -> Parser Int
natural_ n =  int <$> exactly n digit

int :: String -> Int
int =  foldl ((+) . (* 10)) 0 . map (\c -> ord c - ord '0')

-------------------------------------------------------------------------------
-- aanhalingstekens
-------------------------------------------------------------------------------

quoted   :: Parser a -> Parser a
quoted p =  letter '\"' *> p <* letter '\"'

-------------------------------------------------------------------------------
-- regeleinde
-------------------------------------------------------------------------------

linebreak :: Parser [Char]
linebreak =  many1 (letter '\r' <|> letter '\n')

-------------------------------------------------------------------------------
-- interface
-------------------------------------------------------------------------------

just      :: Parser a -> Parser a
just p xs =  [(x, "") | (x, xs') <- p xs, null xs'] 

parse      :: Parser a -> String -> Maybe a
parse p xs =  case just p xs of
  (x, _) : _ -> Just x
  _          -> Nothing